"""himon.schemas package entry file."""
__all__ = ["BaseModel"]

from pydantic import BaseModel as PydanticModel, Extra


class BaseModel(PydanticModel):
    """Base model for himon resources."""

    class Config:
        """Any extra fields will be ignored, strings will have start/end whitespace stripped."""

        anystr_strip_whitespace = True
        allow_population_by_field_name = True
        extra = Extra.ignore



"""
The Exceptions module.

This module provides the following classes:

- ServiceError
- AuthenticationError
"""


class ServiceError(Exception):
    """Class for any API errors."""


class AuthenticationError(ServiceError):
    """Class for any authentication errors."""



"""
The Comic test module.

This module contains tests for Comic objects.
"""
from datetime import date, datetime

from himon.league_of_comic_geeks import LeagueofComicGeeks


def test_comic(session: LeagueofComicGeeks) -> None:
    """Test using the comic endpoint with a valid comic_id."""
    result = session.comic(comic_id=2710631)
    assert result is not None
    assert result.comic_id == 2710631

    assert len(result.characters) == 106
    assert len(result.collected_in) == 13
    assert len(result.creators) == 9
    assert result.date_added == datetime(2012, 7, 2, 23, 15, 17)
    assert result.format == "Comic"
    assert result.is_enabled is True
    assert result.is_nsfw is False
    assert result.is_variant is False
    assert result.isbn is None
    assert len(result.key_events) == 0
    assert result.page_count == 48
    assert result.parent_id is None
    assert result.parent_title is None
    assert result.price == 3.99
    assert result.release_date == date(2009, 7, 15)
    assert result.sku == "MAY090106"
    assert result.title == "Blackest Night #1"
    assert result.upc == 76194128446000111
    assert len(result.variants) == 6


def test_character_null_publisher(session: LeagueofComicGeeks) -> None:
    """Test the comic endpoint with a comic that has a null character publisher name."""
    result = session.comic(comic_id=4174173)
    assert result is not None

    assert result.characters[0].publisher_name is None


def test_key_event_null_universe(session: LeagueofComicGeeks) -> None:
    """Test the comic endpoint with a comic that has a null key event universe name."""
    result = session.comic(comic_id=4174173)
    assert result is not None

    assert result.key_events[0].universe_name is None


def test_upc_alpha_only(session: LeagueofComicGeeks) -> None:
    """Test the comic endpoint with a comic that has an invalid character in the upc."""
    result = session.comic(comic_id=6257084)
    assert result is not None

    assert result.upc is None



"""
The Series test module.

This module contains tests for Series objects.
"""
from datetime import datetime

from himon.league_of_comic_geeks import LeagueofComicGeeks


def test_series(session: LeagueofComicGeeks) -> None:
    """Test using the series endpoint with a valid series_id."""
    result = session.series(series_id=100096)
    assert result is not None
    assert result.series_id == 100096

    assert result.date_added == datetime(2012, 8, 5, 22, 20, 15)
    assert result.is_enabled is True
    assert result.publisher_id == 1
    assert result.publisher_name == "DC Comics"
    assert result.title == "Blackest Night"
    assert result.volume is None
    assert result.year_begin == 2009
    assert result.year_end == 2010



"""
The Exceptions test module.

This module contains tests for Exceptions.
"""
from typing import Optional

import pytest

from himon.exceptions import AuthenticationError, ServiceError
from himon.league_of_comic_geeks import LeagueofComicGeeks


def test_unauthorized() -> None:
    """Test generating an AuthenticationError."""
    session = LeagueofComicGeeks(
        client_id="Invalid",
        client_secret="Invalid",  # noqa: S106
        access_token=None,
        cache=None,
    )
    with pytest.raises(AuthenticationError):
        session.comic(comic_id=1)


def test_not_found(session: LeagueofComicGeeks) -> None:
    """Test a 404 Not Found raises a ServiceError."""
    with pytest.raises(ServiceError):
        session._get_request(endpoint="/invalid")


def test_timeout(client_id: str, client_secret: str, access_token: Optional[str]) -> None:
    """Test a TimeoutError for slow responses."""
    session = LeagueofComicGeeks(
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        timeout=0.1,
        cache=None,
    )
    with pytest.raises(ServiceError):
        session.comic(comic_id=1)



"""
The Comic module.

This module provides the following classes:

- Comic
"""
__all__ = ["Comic"]
from datetime import date, datetime
from typing import Any, Dict, List, Optional

from pydantic import Field, validator

from himon.schemas import BaseModel
from himon.schemas._validators import to_bool, to_optional_float, to_optional_int, to_optional_str
from himon.schemas.search_result import SearchResult
from himon.schemas.series import Series


class Variant(BaseModel):
    """
    The Variant object contains information for a variant comic.

    Attributes:
        date_modified: Date and time when the Variant was last updated.
        price: Price of the Variant.
        release_date: The date the Variant was released.
        title: Name/Title of the Variant.
        variant_id: Identifier used by League of Comic Geeks.
    """

    date_modified: datetime
    price: Optional[float] = None
    release_date: date = Field(alias="date_release")
    title: str
    variant_id: int = Field(alias="id")

    @validator("price", pre=True)
    def _to_optional_float(cls, v: str) -> Optional[float]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_float(v)


class KeyEvent(BaseModel):
    """
    The KeyEvent object contains information for a key event.

    Attributes:
        character_id: Identifier used by League of Comic Geeks.
        event_id: Identifier used by League of Comic Geeks.
        name: Name/Title of the Event.
        note: Unknown field
        parent_name: Unknown field
        type: Unknown field
        type_id: Identifier used by League of Comic Geeks.
        universe_name: Universe this Event took place in.
    """

    character_id: int
    event_id: int = Field(alias="id")
    name: str
    note: Optional[str] = None
    parent_name: Optional[str] = None
    type: int  # How is it different to type_id?
    type_id: int
    universe_name: Optional[str] = None

    @validator("note", "parent_name", pre=True)
    def _to_optional_str(cls, v: str) -> Optional[str]:
        """Pydantic validator to convert a Str to None or return html stripped value."""
        return to_optional_str(v)


class Creator(BaseModel):
    """
    The Creator object contains information for a creator.

    Attributes:
        creator_id: Identifier used by League of Comic Geeks.
        name: Name/Title of the Creator.
        role: List of roles the Creator has in the comic. Separated by `,`.
        role_id: List of role ids the Creator has in the comic. Separated by `,`.
    """

    creator_id: int = Field(alias="id")
    name: str
    role: str
    role_id: str

    @property
    def roles(self) -> Dict[int, str]:
        """Return a dict of role id and role name the Creator is attached to."""
        role_dict = {}
        id_list = self.role_id.split(",")
        role_list = self.role.split(",")
        for index, role_id in enumerate(id_list):
            role_dict[int(role_id)] = role_list[index].strip().title()
        return role_dict


class Character(BaseModel):
    """
    The Character object contains information for a character.

    Attributes:
        character_id: Identifier used by League of Comic Geeks.
        date_added: Date and time when the Character was added.
        date_modified: Date and time when the Character was last updated.
        full_name: Full name of Character
        is_enabled: Unknown field
        name: Name/Alias of Character.
        parent_id: Unknown field
        parent_name: Unknown field
        publisher_name: The publisher name of Character.
        type_id: Unknown field
        universe_id: Universe id this Character is from.
        universe_name: Universe name this Character is from.
    """

    character_id: int = Field(alias="id")
    date_added: datetime
    date_modified: datetime
    full_name: str
    is_enabled: bool = Field(alias="enabled")
    name: str
    parent_id: Optional[int] = None
    parent_name: Optional[str] = None
    publisher_name: Optional[str] = None
    type_id: int
    universe_id: Optional[int] = None
    universe_name: Optional[str] = None

    @validator("parent_name", "universe_name", pre=True)
    def _to_optional_str(cls, v: str) -> Optional[str]:
        """Pydantic validator to convert a Str to None or return html stripped value."""
        return to_optional_str(v)

    @validator("parent_id", "universe_id", pre=True)
    def _to_optional_int(cls, v: str) -> Optional[int]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_int(v)

    @validator("is_enabled", pre=True)
    def _to_bool(cls, v: str) -> bool:
        """Pydantic validator to convert a Str 0/1 to a bool."""
        return to_bool(v)


class Comic(BaseModel):
    """
    The Comic object contains information for a comic.

    Attributes:
        characters: List of Characters in the Comic.
        collected_in: List of Comics this has been collected in.
        comic_id: Identifier used by League of Comic Geeks.
        creators: List of Creators associated with the Comic
        date_added: Date and time when the Comic was added.
        date_modified: Date and time when the Comic was last updated.
        description: Description of the Comic.
        format: Type of Comic.
        is_enabled: Unknown field
        is_nsfw: Comic has been marked as NSFW
        is_variant: Comic has been marked as Variant
        isbn: ISBN identifier
        key_events: List of Key Events taken place in the Comic.
        page_count: Count of pages in the Comic.
        parent_id: If it is a variant comic, id of the original comic.
        parent_title: If it is a variant comic, title of the origin comic.
        price: Price of Comic
        release_date: The date the Comic was released.
        series: The series this Comic comes from.
        sku: SKU identifier.
        title: Name/Title of the Comic.
        upc: UPC identifier.
        variants: List of variants this Comic has.
    """

    characters: List[Character] = Field(default_factory=list)
    collected_in: List[SearchResult] = Field(default_factory=list)
    comic_id: int = Field(alias="id")
    creators: List[Creator] = Field(default_factory=list)
    date_added: datetime
    date_modified: datetime
    description: Optional[str] = None
    format: str
    is_enabled: bool = Field(alias="enabled")
    is_nsfw: bool = Field(alias="nsfw")
    is_variant: bool = Field(alias="variant")
    isbn: Optional[int] = None
    key_events: List[KeyEvent] = Field(default_factory=list)
    page_count: int = Field(alias="pages")
    parent_id: Optional[int] = None
    parent_title: Optional[str] = None
    price: Optional[float] = None
    release_date: date = Field(alias="date_release")
    series: Series
    sku: str
    title: str
    upc: Optional[int] = None
    variants: List[Variant] = Field(default_factory=list)

    def __init__(self, **data: Any):
        if data["keys"]:
            data["key_events"] = list(data["keys"].values())
        for key, value in data["details"].items():
            data[key] = value
        super().__init__(**data)

    @validator("isbn", "parent_id", "upc", pre=True)
    def _to_optional_int(cls, v: str) -> Optional[int]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_int(v)

    @validator("is_enabled", "is_nsfw", "is_variant", pre=True)
    def _to_bool(cls, v: str) -> bool:
        """Pydantic validator to convert a Str 0/1 to a bool."""
        return to_bool(v)

    @validator("description", "parent_title", pre=True)
    def _to_optional_str(cls, v: str) -> Optional[str]:
        """Pydantic validator to convert a Str to None or return html stripped value."""
        return to_optional_str(v)

    @validator("price", pre=True)
    def _to_optional_float(cls, v: str) -> Optional[float]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_float(v)



"""
The conftest module.

This module contains pytest fixtures.
"""
import os
from typing import Optional

import pytest

from himon.league_of_comic_geeks import LeagueofComicGeeks
from himon.sqlite_cache import SQLiteCache


@pytest.fixture(scope="session")
def client_id() -> str:
    """Retrieve the Client Id from environment variables."""
    return os.getenv("LEAGUE_OF_COMIC_GEEKS__CLIENT_ID", default="IGNORED")


@pytest.fixture(scope="session")
def client_secret() -> str:
    """Retrieve the Client Secret from environment variables."""
    return os.getenv("LEAGUE_OF_COMIC_GEEKS__CLIENT_SECRET", default="IGNORED")


@pytest.fixture(scope="session")
def access_token() -> Optional[str]:
    """Retrieve the Access Token from environment variables."""
    return os.getenv("LEAGUE_OF_COMIC_GEEKS__ACCESS_TOKEN")


@pytest.fixture(scope="session")
def session(client_id: str, client_secret: str, access_token: Optional[str]) -> LeagueofComicGeeks:
    """Set the Himon session fixture."""
    service = LeagueofComicGeeks(
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        cache=SQLiteCache("tests/cache.sqlite", expiry=None),
    )
    if not access_token:
        service.access_token = service.generate_access_token()
    return service



"""
The League of Comic Geeks module.

This module provides the following classes:

- LeagueofComicGeeks
"""
import platform
from typing import Any, Dict, List, Optional
from urllib.parse import urlencode

from pydantic import ValidationError, parse_obj_as
from ratelimit import limits, sleep_and_retry
from requests import get
from requests.exceptions import ConnectionError, HTTPError, JSONDecodeError, ReadTimeout

from himon import __version__
from himon.exceptions import AuthenticationError, ServiceError
from himon.schemas.comic import Comic
from himon.schemas.search_result import SearchResult
from himon.schemas.series import Series
from himon.sqlite_cache import SQLiteCache

MINUTE = 60


class LeagueofComicGeeks:
    """
    Wrapper to allow calling League of Comic Geeks API endpoints.

    Args:
        client_id: User's Client Id to access League of Comic Geeks.
        client_secret: User's Client Secret to access League of Comic Geeks.
        access_token: User's Access Token to access League of Comic Geeks.
        timeout: Set how long requests will wait for a response (in seconds).
        cache: SQLiteCache to use if set.

    Attributes:
        headers (Dict[str, str]): Header used when requesting from League of Comic Geeks.
        timeout (int): How long requests will wait for a response (in seconds).
        cache (Optional[SQLiteCache]): SQLiteCache to use if set.
        client_secret (str): User's Client Secret to access League of Comic Geeks.
        access_token (Optional[str]): User's Access Token to access League of Comic Geeks.
    """

    API_URL = "https://leagueofcomicgeeks.com/api"

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        access_token: Optional[str] = None,
        timeout: int = 30,
        cache: Optional[SQLiteCache] = None,
    ):
        self.headers = {
            "Accept": "application/json",
            "User-Agent": f"Himon/{__version__}/{platform.system()}: {platform.release()}",
            "X-API-CLIENT": client_id,
        }
        self.timeout = timeout
        self.cache = cache

        self.client_secret = client_secret
        self.access_token = access_token

    @sleep_and_retry
    @limits(calls=20, period=MINUTE)
    def _perform_get_request(self, url: str, params: Dict[str, str] = None) -> Dict[str, Any]:
        """
        Make GET request to League of Comic Geeks.

        Args:
            url: The url to request information from.
            params: Parameters to add to the request.

        Returns:
            Json response from League of Comic Geeks.

        Raises:
            ServiceError: If there is an issue with the request or response.
            AuthenticationError:
                If League of Comic Geeks returns with an invalid API Key or Client Id response.
        """
        if params is None:
            params = {}

        try:
            response = get(url, params=params, headers=self.headers, timeout=self.timeout)
            response.raise_for_status()
            return response.json()
        except ConnectionError as err:
            raise ServiceError(f"Unable to connect to `{url}`") from err
        except HTTPError as err:
            if err.response.status_code == 403:
                raise AuthenticationError("Invalid API Key") from err
            if err.response.status_code == 404:
                raise ServiceError("Unknown endpoint") from err
            raise ServiceError(f"{err.response.status_code}: {err.response.text}") from err
        except JSONDecodeError as err:
            raise ServiceError(f"Unable to parse response from `{url}` as Json") from err
        except ReadTimeout as err:
            raise ServiceError("Service took too long to respond") from err

    def _get_request(
        self,
        endpoint: str,
        params: Dict[str, str] = None,
        skip_cache: bool = False,
    ) -> Dict[str, Any]:
        """
        Check cache or make GET request to League of Comic Geeks.

        Args:
            endpoint: The endpoint to request information from.
            params: Parameters to add to the request.
            skip_cache: Don't save or read from the cache.

        Returns:
            Json response from League of Comic Geeks.

        Raises:
            ServiceError: If there is an issue with the request or response.
            AuthenticationError:
                If League of Comic Geeks returns with an invalid API Key or Client Id response.
        """
        if params is None:
            params = {}

        url = self.API_URL + endpoint
        cache_params = f"?{urlencode(params)}" if params else ""
        cache_key = f"{url}{cache_params}"

        if self.cache and not skip_cache and (cached_response := self.cache.select(cache_key)):
            return cached_response

        response = self._perform_get_request(url=url, params=params)

        if self.cache and not skip_cache:
            self.cache.insert(cache_key, response)

        return response

    @sleep_and_retry
    @limits(calls=20, period=MINUTE)
    def _str_get_request(self, endpoint: str, params: Dict[str, str] = None) -> str:
        """
        Make GET request to League of Comic Geeks, expecting a str response.

        Args:
            endpoint: The endpoint to request information from.
            params: Parameters to add to the request.

        Returns:
            String response from League of Comic Geeks.

        Raises:
            ServiceError: If there is an issue with the request or response.
            AuthenticationError:
                If League of Comic Geeks returns with an invalid API Key or Client Id response.
        """
        if params is None:
            params = {}

        url = self.API_URL + endpoint

        try:
            response = get(url, params=params, headers=self.headers, timeout=self.timeout)
            response.raise_for_status()
            return response.json()
        except ConnectionError as err:
            raise ServiceError(f"Unable to connect to `{url}`") from err
        except HTTPError as err:
            if err.response.status_code == 403:
                raise AuthenticationError("Invalid API Key") from err
            if err.response.status_code == 404:
                raise ServiceError("Unknown endpoint") from err
            raise ServiceError(f"{err.response.status_code}: {err.response.text}") from err
        except JSONDecodeError as err:
            raise ServiceError(f"Unable to parse response from `{url}` as Json") from err
        except ReadTimeout as err:
            raise ServiceError("Service took too long to respond") from err

    def generate_access_token(self) -> str:
        """
        Request an access token.

        Returns:
            An access token.

        Raises:
            ServiceError: If there is an issue with the client id or secret.
        """
        self.headers["X-API-KEY"] = self.client_secret
        return self._str_get_request("/authorize/format/json")

    def search(self, search_term: str) -> List[SearchResult]:
        """
        Request a list of search results.

        Args:
            search_term: Search query string
        Returns:
            A list of results.

        Raises:
            ServiceError: If there is an issue with validating the response.
        """
        try:
            self.headers["X-API-KEY"] = self.access_token
            results = self._get_request("/search/format/json", params={"query": search_term})
            return parse_obj_as(List[SearchResult], results)
        except ValidationError as err:
            raise ServiceError(err) from err

    def get_series(self, series_id: int) -> Series:
        """
        Request data for a Series based on its id.

        Args:
            series_id: The Series id.

        Returns:
            A Series object.

        Raises:
            ServiceError: If there is an issue with validating the response.
        """
        try:
            self.headers["X-API-KEY"] = self.access_token
            result = self._get_request(
                "/series/format/json",
                params={"series_id": str(series_id)},
            )
            if "details" in result:
                result = result["details"]
            return parse_obj_as(Series, result)
        except ValidationError as err:
            raise ServiceError(err) from err

    def series(self, series_id: int) -> Series:
        """
        Request data for a Series based on its id.

        **DEPRECATED**: Use get_series()

        Args:
            series_id: The Series id.

        Returns:
            A Series object.

        Raises:
            ServiceError: If there is an issue with validating the response.
        """
        return self.get_series(series_id=series_id)

    def get_comic(self, comic_id: int) -> Comic:
        """
        Request data for a Comic based on its id.

        Args:
            comic_id: The Comic id.

        Returns:
            A Comic object.

        Raises:
            ServiceError: If there is an issue with validating the response.
        """
        try:
            self.headers["X-API-KEY"] = self.access_token
            result = self._get_request(
                "/comic/format/json",
                params={"comic_id": str(comic_id)},
            )
            return parse_obj_as(Comic, result)
        except ValidationError as err:
            raise ServiceError(err) from err

    def comic(self, comic_id: int) -> Comic:
        """
        Request data for a Comic based on its id.

        **DEPRECATED**: Use get_comic()

        Args:
            comic_id: The Comic id.

        Returns:
            A Comic object.

        Raises:
            ServiceError: If there is an issue with validating the response.
        """
        return self.get_comic(comic_id=comic_id)



"""himon package entry file."""
__version__ = "0.4.0"
__all__ = ["__version__", "get_cache_root"]

import os
from pathlib import Path


def get_cache_root() -> Path:
    """
    Create and return the path to the cache for Himon, supports XDG_CACHE_HOME env.

    Returns:
        The path to the Himon cache folder.
    """
    cache_home = os.getenv("XDG_CACHE_HOME", default=str(Path.home() / ".cache"))
    folder = Path(cache_home).resolve() / "himon"
    folder.mkdir(parents=True, exist_ok=True)
    return folder



"""
The Search Result module.

This module provides the following classes:

- SearchResult
"""
__all__ = ["SearchResult"]
from datetime import date, datetime
from typing import Optional

from pydantic import Field, validator

from himon.schemas import BaseModel
from himon.schemas._validators import to_bool, to_optional_float, to_optional_int, to_optional_str


class SearchResult(BaseModel):
    """
    The SearchResult object contains information for a search result.

    Attributes:
        comic_id: Identifier used by League of Comic Geeks.
        date_modified: Date and time when the Comic was last updated.
        description: Description of the Comic.
        format: Type of Comic.
        is_enabled: Unknown field
        is_variant: Comic has been marked as Variant.
        parent_id: If it is a variant comic, id of the original comic.
        parent_title: If it is a variant comic, title of the original comic.
        price: Price of Comic.
        publisher_id: The publisher id of the Comic.
        publisher_name: The publisher name of the Comic.
        release_date: The date the Comic was released.
        series_id: The series id of the Comic.
        series_name: The series name of the Comic.
        series_volume: The series volume of the Comic.
        title: Name/Title of the Comic.
        year_begin: The year the Series started.
        year_end: The year the Series ended.
    """

    comic_id: int = Field(alias="id")
    date_modified: datetime
    description: Optional[str] = None
    format: str
    is_enabled: bool = Field(alias="enabled")
    is_variant: bool = Field(alias="variant")
    parent_id: Optional[int] = None
    parent_title: Optional[str] = None
    price: Optional[float] = None
    publisher_id: int
    publisher_name: str
    release_date: date = Field(alias="date_release")
    series_id: int
    series_name: str
    series_volume: Optional[int] = None
    title: str
    year_begin: int = Field(alias="series_begin")
    year_end: Optional[int] = Field(alias="series_end", default=None)

    @validator("parent_id", "series_volume", "year_end", pre=True)
    def _to_optional_int(cls, v: str) -> Optional[int]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_int(v)

    @validator("is_variant", "is_enabled", pre=True)
    def _to_bool(cls, v: str) -> bool:
        """Pydantic validator to convert a Str 0/1 to a bool."""
        return to_bool(v)

    @validator("price", pre=True)
    def _to_optional_float(cls, v: str) -> Optional[float]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_float(v)

    @validator("description", "parent_title", pre=True)
    def _to_optional_str(cls, v: str) -> Optional[str]:
        """Pydantic validator to convert a Str to None or return html stripped value."""
        return to_optional_str(v)



"""
The Series module.

This module provides the following classes:

- Series
"""
__all__ = ["Series"]
from datetime import datetime
from typing import Optional

from pydantic import Field, validator

from himon.schemas import BaseModel
from himon.schemas._validators import to_bool, to_optional_int, to_optional_str


class Series(BaseModel):
    """
    The Series object contains information for a series.

    Attributes:
        date_added: Date and time when the Series was added.
        date_modified: Date and time when the Series was last updated.
        description: Description of the Series.
        is_enabled: Unknown field
        publisher_id: The publisher id of the Series.
        publisher_name: The publisher name of the Series.
        series_id: Identifier used by League of Comic Geeks.
        title: Name/Title of the Series.
        volume: Volume number.
        year_begin: The year the Series started.
        year_end: The year the Series ended.
    """

    date_added: datetime
    date_modified: datetime
    description: Optional[str] = None
    is_enabled: bool = Field(alias="enabled")
    publisher_id: int
    publisher_name: str
    series_id: int = Field(alias="id")
    title: str
    volume: Optional[int] = None
    year_begin: int
    year_end: Optional[int] = None

    @validator("description", pre=True)
    def _to_optional_str(cls, v: str) -> Optional[str]:
        """Pydantic validator to convert a Str to None or return html stripped value."""
        return to_optional_str(v)

    @validator("is_enabled", pre=True)
    def _to_bool(cls, v: str) -> bool:
        """Pydantic validator to convert a Str 0/1 to a bool."""
        return to_bool(v)

    @validator("volume", "year_end", pre=True)
    def _to_optional_int(cls, v: str) -> Optional[int]:
        """Pydantic validator to convert a Str or 0 to None or return value."""
        return to_optional_int(v)



__all__ = ["to_optional_str", "to_optional_int", "to_optional_float", "to_bool"]

import html
import re
from typing import Optional


def to_optional_int(v: str) -> Optional[int]:
    """
    Convert a Str or 0 to None or return value.

    Args:
        v: Value to convert
    Return:
        Value mapped as None or value
    """
    try:
        if not v or int(v) == 0:
            return None
        return int(v)
    except ValueError:
        return None


def to_optional_float(v: str) -> Optional[float]:
    """
    Convert a Str or 0 to None or return value.

    Args:
        v: Value to convert
    Return:
        Value mapped as None or value
    """
    if v:
        v = str(v).replace("..", ".")
    try:
        if not v or float(v) == 0:
            return None
        return float(v)
    except ValueError:
        return None


def to_bool(v: str) -> bool:
    """
    Convert a Str 0/1 to a bool.

    Args:
        v: Value to convert
    Return:
        Value mapped as Bool
    Raises:
        ValueError: If value isn't a 0/1
    """
    if str(v) == "0":
        return False
    if str(v) == "1":
        return True
    raise ValueError(f"Unknown bool value `{v}`.")


def to_optional_str(v: str) -> Optional[str]:
    """
    Convert a Str to None or return html stripped value.

    Args:
        v: Value to convert
    Return:
        Value mapped as None or html stripped value
    """
    if not v:
        return None
    regex = re.compile(r"(<!--.*?-->|<[^>]*>)")
    return " ".join(html.unescape(regex.sub("", v.strip())).split())



"""
The SearchResult test module.

This module contains tests for SearchResult objects.
"""
from datetime import date

from himon.league_of_comic_geeks import LeagueofComicGeeks


def test_search_result(session: LeagueofComicGeeks) -> None:
    """Test using the search endpoint with a valid comic title."""
    results = session.search(search_term="Blackest Night #1")
    assert len(results) != 0
    result = [x for x in results if x.comic_id == 2710631][0]
    assert result is not None
    assert result.comic_id == 2710631

    assert result.format == "Comic"
    assert result.is_enabled is True
    assert result.is_variant is False
    assert result.parent_id is None
    assert result.parent_title is None
    assert result.price == 3.99
    assert result.publisher_id == 1
    assert result.publisher_name == "DC Comics"
    assert result.release_date == date(2009, 7, 15)
    assert result.series_id == 100096
    assert result.series_name == "Blackest Night"
    assert result.series_volume is None
    assert result.title == "Blackest Night #1"
    assert result.year_begin == 2009
    assert result.year_end == 2010



"""
The SQLiteCache module.

This module provides the following classes:

- SQLiteCache
"""
__all__ = ["SQLiteCache"]
import json
import sqlite3
from datetime import date, timedelta
from pathlib import Path
from typing import Any, Dict, Optional

from himon import get_cache_root


class SQLiteCache:
    """
    The SQLiteCache object to cache search results from League of Comic Geeks.

    Args:
        path: Path to database.
        expiry: How long to keep cache results.

    Attributes:
        expiry (Optional[int]): How long to keep cache results.
        con (sqlite3.Connection): Database connection
    """

    def __init__(
        self,
        path: Path = None,
        expiry: Optional[int] = 14,
    ):
        self.expiry = expiry
        self.con = sqlite3.connect(path or get_cache_root() / "cache.sqlite")
        self.con.row_factory = sqlite3.Row

        self.con.execute("CREATE TABLE IF NOT EXISTS queries (query, response, query_date);")
        self.delete()

    def select(self, query: str) -> Dict[str, Any]:
        """
        Retrieve data from the cache database.

        Args:
            query: Search string
        Returns:
            Empty dict or select results.
        """
        if self.expiry:
            expiry = date.today() - timedelta(days=self.expiry)
            cursor = self.con.execute(
                "SELECT * FROM queries WHERE query = ? and query_date > ?;",
                (query, expiry.isoformat()),
            )
        else:
            cursor = self.con.execute("SELECT * FROM queries WHERE query = ?;", (query,))
        if results := cursor.fetchone():
            return json.loads(results["response"])
        return {}

    def insert(self, query: str, response: Dict[str, Any]) -> None:
        """
        Insert data into the cache database.

        Args:
            query: Search string
            response: Data to save
        """
        self.con.execute(
            "INSERT INTO queries (query, response, query_date) VALUES (?, ?, ?);",
            (query, json.dumps(response), date.today().isoformat()),
        )
        self.con.commit()

    def delete(self) -> None:
        """Remove all expired data from the cache database."""
        if not self.expiry:
            return
        expiry = date.today() - timedelta(days=self.expiry)
        self.con.execute("DELETE FROM queries WHERE query_date < ?;", (expiry.isoformat(),))
        self.con.commit()
